# coding: utf-8
"""mq 相关的工具类"""

import datetime
import logging
import time

import pika
import pika.exceptions
from pika import BasicProperties

from . import settings
from .tools import generate_uid

log = logging.getLogger(__name__)


def _new_connection():
    params = pika.URLParameters(settings.MQ_URL)
    return pika.BlockingConnection(parameters=params)


class RpcClient(object):
    def __init__(self, rpc_queue='rpc_queue', save=True):
        # self.connection = get_connection()
        # self.channel = get_channel()
        self.__init_connection()
        self.response = None
        self.corr_id = None
        self.rpc_queue = rpc_queue

        self.save = save

    def __del__(self):
        """todo 对象回收的逻辑。"""
        # log.info('RpcClient instance is destroyed: [%s] %s', self.rpc_queue, self.callback_queue)

    def __init_connection(self):
        if hasattr(self, 'channel') and not self.channel.is_closed:
            try:
                self.channel.close()
            except:
                pass
        self.connection = _new_connection()
        self.channel = self.connection.channel()
        result = self.channel.queue_declare(queue='', exclusive=True)
        self.callback_queue = result.method.queue
        self.channel.basic_consume(
            queue=self.callback_queue,
            on_message_callback=self.on_response,
            auto_ack=True)

    def on_response(self, ch, method, props, body):
        log.info('callback: %s, %s, %s, %s', ch, method, props, body[:100])
        if self.corr_id == props.correlation_id:
            if self.save:
                from . import fileman
                fileman.save_file_from_bytes(body,
                                             f'rpcclient/response/{datetime.datetime.today().strftime("%Y%m%d")}/{props.correlation_id}.data')
            self.response = body
        else:
            log.warning('not correct relation id: %s <> %s', self.corr_id, props.correlation_id)

    def call(self, byte_content: bytes, timeout=30, extra_headers=None):
        self.response = None
        self.corr_id = generate_uid()

        log.info('start call: rpc_queue=%s, corr_id=%s, timeout=%s, cb_queue=%s', self.rpc_queue, self.corr_id, timeout,
                 self.callback_queue)
        try:
            self.channel.basic_publish(
                exchange='',
                routing_key=self.rpc_queue,
                properties=pika.BasicProperties(
                    reply_to=self.callback_queue,
                    correlation_id=self.corr_id,
                    headers=extra_headers,
                    expiration=str(timeout * 1000),
                ),
                body=byte_content)
        except:
            # 失败则重新建立连接，重试一次
            log.exception('连接丢失，重试')
            self.__init_connection()
            self.channel.basic_publish(
                exchange='',
                routing_key=self.rpc_queue,
                properties=pika.BasicProperties(
                    reply_to=self.callback_queue,
                    correlation_id=self.corr_id,
                    headers=extra_headers,
                    expiration=str(timeout * 1000),
                ),
                body=byte_content)
        start = time.time()
        while self.response is None:
            log.debug('self.response is None in rpcclient')
            self.connection.process_data_events(2)
            if time.time() - start > timeout:
                raise Exception('超时, 队列没有数据返回: reply_to=%s, correlation_id=%s' % (self.callback_queue, self.corr_id))
        return self.response


class RpcServer(object):

    def __init__(self, function_callback, queue='rpc_queue'):
        """
        function_callback: 处理 body(bytes) 的函数体。这个函数的返回值应该是 bytes，会直接返回给客户端
        """
        self.function_callback = function_callback
        self.queue = queue

    def on_request(self, ch, method, props, body):
        log.info('Rpc Server[%s][%s] received request: %s', self.queue, self.function_callback,
                 props.correlation_id)
        try:
            response = self.function_callback(body)
        except Exception as e:
            response = str(e)

        ch.basic_publish(exchange='',
                         routing_key=props.reply_to,
                         properties=pika.BasicProperties(correlation_id= \
                                                             props.correlation_id),
                         body=response)
        ch.basic_ack(delivery_tag=method.delivery_tag)

    def start_consuming(self):
        self.connection = _new_connection()
        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(queue=self.queue, on_message_callback=self.on_request)
        self.channel.start_consuming()


def publish(exchange: str, routing_key: str, body: bytes, properties: BasicProperties = None, mandatory: bool = False):
    _new_connection().channel().basic_publish(exchange, routing_key, body, properties, mandatory)


def server_reply(channel, props, response):
    if props.reply_to:
        channel.basic_publish(exchange='',
                              routing_key=props.reply_to,
                              properties=pika.BasicProperties(
                                  correlation_id=props.correlation_id
                              ),
                              body=response)


class Consumer(object):
    class ConsumerExitError(Exception):
        pass

    def __init__(self, on_message, queue, prefetch_count=1):
        """注意 queue 一定要事先创建好"""
        self.on_message = on_message
        self.prefetch_count = prefetch_count
        self.queue = queue
        self.channel = None
        self.connection = None

    def start_consuming(self):
        log.info('Consumer is consuming now: %s', self.queue)
        while True:
            try:
                url_param = settings.MQ_URL
                params = pika.URLParameters(url_param)
                self.connection = pika.BlockingConnection(parameters=params)
                self.channel = self.connection.channel()
                self.channel.basic_qos(prefetch_count=self.prefetch_count)
                self.channel.basic_consume(queue=self.queue, on_message_callback=self.on_message)
                try:
                    self.channel.start_consuming()
                except Consumer.ConsumerExitError:
                    log.warning('Consumer is exit now: %s', self.queue)
                    break
                except KeyboardInterrupt:
                    self.channel.stop_consuming()
                    self.connection.close()
                    break
            except pika.exceptions.ConnectionClosedByBroker:
                # Uncomment this to make the example not attempt recovery
                # from server-initiated connection closure, including
                # when the node is stopped cleanly
                #
                # break
                log.warning('pika connection closed. Reconnecting...')
                continue
                # Do not recover on channel errors
            except pika.exceptions.AMQPChannelError as err:
                log.error("Caught a channel error: %s, stopping...", err)
                break
                # Recover on all other connection errors
            except pika.exceptions.AMQPConnectionError:
                log.warning('Connection was closed, retrying...')
                continue
